;;;;SMSYNTAX.SCM   Code for 6.001 Spring 95
;;; Syntax for Substitution Model

;;TAG TEST

(define (tagged-pair? tag)
  (lambda (exp)
    (and (pair? exp)
         (eq? (car exp) tag))))


;;DEFINE

(define (make-define var exp)
  (list 'define var exp))

(define define? (tagged-pair? 'define))

(define (variable-of-define def)
  ;;handles both (define f exp) and (define (f x1 ...) body)
  (let ((var-form (cadr def)))
    (if (variable? var-form)
        var-form
        (car var-form))))
      
(define (expression-of-define def)
  ;;handles both (define f exp) and (define (f x1 ...) body)
  (let ((var-form (cadr def))) 
    (if (variable? var-form)
        (caddr def)
        (make-lambda
         (cdr var-form)               ;formals
         (let ((body-parts (cddr def)))
           (if (body-with-defines? body-parts)
               body-parts
               (car body-parts))))))) ;BODY-PARTS must be length 1 list of <exp>


;;EXPRESSION

(define (expression? body)
  (or (self-evaluating? body)
      (variable? body)
      (special-form? body)
      (symbol-expression? body)       ;redundant in our representation
      (submodel-null? body)           ;redundant in our representation
      (combination? body)))

(define (special-form? body)
  (and (pair? body) (expression-keyword? (car body))))

;;BODY

(define (body-with-defines? body)
  (and (list? body)
       (pair? body)
       (pair? (cdr body))
       (expression? (last-pair body))
       (for-all? (except-last-pair body) define?)))

(define (make-body defines exp)
 (if (null? defines)
     exp
     (append defines (list exp))))

(define (defines-of-body body)
  (if (expression? body)
      '()
      (except-last-pair body)))

(define (expression-of-body body)
  (if (expression? body)
      body
      (car (last-pair body))))

;;SELF-EVALUATING

(define (self-evaluating? body)
  (or (boolean? body)
      (number? body)
      (string? body)))

;;SYMBOL

(define (symbol-expression? body)
  (and ((tagged-pair? 'quote) body)
       (pair? (cdr body))
       (symbol? (cadr body))
       (null? (cddr body))))

(define (make-symbol name)
  (list 'quote name))

(define symbol-name cadr)

;;COMBINATION

(define (make-combination exps) exps)

(define (expressions-of-combination comb) comb)

(define (combination? body)
  (and (pair? body)
       (list? body)
       (for-all? body expression?)))

(define (operator-of comb) (car (expressions-of-combination comb)))
(define (operands-of comb) (cdr (expressions-of-combination comb)))

;;LAMBDA-EXPRESSION

(define (make-lambda formals body)
  (cons 'lambda
        (cons formals
              (if (expression? body)
                  (list body)
                  body))))
(define lambda-expression? (tagged-pair? 'lambda))
(define formals-of-lambda cadr)
(define (body-of-lambda lam-exp)
  (let ((bodylist (cddr lam-exp)))
    (if (null? (cdr bodylist))          ;more efficient than (= (length bodylist) 1)
        (car bodylist)
        bodylist)))                 

;;IF

(define (make-if test consequent alternative)
  (list 'if test consequent alternative))

(define if? (tagged-pair? 'if))
  
(define test-of-if cadr)
(define consequent-of caddr)
(define (no-alternative? ifexp)
  (null? (cdddr ifexp)))
(define alternative-of cadddr)

;;VARIABLE

(define (variable? body)
  (and (symbol? body)
       (not (keyword? body))))

(define (make-variable name) name)
(define (variable-name var) var)

(define equ-variables? eq?)

(define (primitive-procedure-variable? body)
  (and (symbol? body)
       (memq body primitive-procedure-variables)))

(define primitive-procedure-variables
  '(+ - * / = < > >= <= inc dec 1+ -1+ 
    quotient sqrt expt round abs gcd
    max min remainder modulo zero? 
    positive? negative? odd? even?
    exp log sin cos tan asin acos atan
    number? not boolean? length list? string?
    symbol? eq? procedure?))

(define (rule-specified-procedure-variable? var)
  (and (symbol? var)
       (assq var rule-association-list)))


;;SET!

(define set!? (tagged-pair? 'set!))

(define (make-set! variable value)
  (list 'set! variable value))

(define variable-of-set! cadr)
(define expression-of-set! caddr)


;;BEGIN

(define begin? (tagged-pair? 'begin))

(define (make-begin exps)
  (cons 'begin exps))

(define expressions-of-begin cdr)


;;KEYWORD

(define expression-keywords		;AND, OR, COND, LET treated as sugar
  '(quote lambda if let cond set! begin and or))

(define nonexpression-keywords '(define else))

(define keywords (append nonexpression-keywords expression-keywords))

(define (keyword? tag)
  (and (symbol? tag)
       (memq tag keywords)))

(define (expression-keyword? tag)
  (and (symbol? tag)
       (memq tag expression-keywords)))

;;VALUE

(define (immediate-value? body)
  (or (self-evaluating? body)
      (submodel-null? body)
      (lambda-expression? body)
      (symbol-expression? body)
      (and (combination? body)
           (eq? 'cons (operator-of body))
           (let ((rands (operands-of body)))
             (and (immediate-value? (car rands))
                  (immediate-value? (cadr rands)))))
      (primitive-procedure-variable? body)
      (rule-specified-procedure-variable? body)))


(define false-value? not)

(define submodel-nil '())
(define submodel-null? null?)

;(define submodel-nil '(quote ()))

;(define (submodel-null? body)
;  (and
;   ((tagged-pair? 'quote) body)
;   (null? (cadr body))
;   (null? (cddr body))))

;;;in the Sub Model, the USELESS-VALUE will be a symbol

(define submodel-useless-value (make-symbol 'useless-value))
(define (submodel-useless-value? body) (equal? body submodel-useless-value))

;;TAGGED TYPES

;;selector TAG-OF returns the tag of any tagged type
(define tag-of car)

;; Tagged-Body  =  Labelled-Sum[stepped: <body>,
;;                                 set!: Set!-info X <exp>,
;;                                  val: <body>,
;;                                stuck: Error-info]
;; Set!-info  =  Var X <immediate value>
;; Body-Tag = {stepped, set!, val, stuck}

;;constructor
;;MAKE-TAGGED-BODY: (Body-Tag,(<body> U Error-Info U (Set!-info X <exp>)))
;;                                                    --> Tagged-Body

;;selector
;;DATUM-OF-TAGGED-BODY: Tagged-Body
;;                         --> (<body> U Error-Info U (Set!-info X <exp>))

(define make-tagged-body list)
(define datum-of-tagged-body cadr)


;; Tagged-Exps  =  Labelled-Sum[stepped: List(<define>) X List(<exp>),
;;                                 set!: Set!-info X ({'()} X List(<exp>)),
;;                                  val: Empty,
;;                                stuck: Error-info]

(define make-tagged-exps list)
(define datum-of-tagged-exps cdr)

         ;The following only work for 'stepped.  Another cdr is needed for 'set!.
         ;Should have made separate selectors for different tags.
(define expressions-of-exps-datum cadr)
(define defines-of-exps-datum car)


;; Tagged-Defs   =  Labelled-Sum[stepped: List(<define>),
;;                                 set!: Set!-info X ({'()} X List(<define>)),
;;                                   val: Empty,
;;                                 stuck: Error-info]

         ;The following only work for 'stepped.  Another cdr is needed for 'set!.
         ;Should have made separate selectors for different tags.
(define make-tagged-defs list)
(define defines-of-defs-datum cadr)


;;;;DESUGAR


;;AND
(define and? (tagged-pair? 'and))
(define expressions-of-and cdr)
(define (make-and exps) (cons 'and exps))


;;OR
(define or? (tagged-pair? 'or))
(define expressions-of-or cdr)
(define (make-or exps) (cons 'or exps))


;;QUOTE
(define quote? (tagged-pair? 'quote))
(define text-of-quote cadr)
(define (make-quote text) (list 'quote text))


;;COND
(define cond? (tagged-pair? 'cond))
(define clauses-of-cond cdr)
(define test-of-clause car)
(define expression-of-clause cadr)
(define else-clause? (tagged-pair? 'else))
(define (make-cond clauses) (cons 'cond clauses))

;;LET

(define let? (tagged-pair? 'let))
(define bindings-of-let cadr)
(define variable-of-binding car)
(define init-of-binding cadr)
(define (body-of-let let-exp)
  (let ((bodylist (cddr let-exp)))
    (if (null? (cdr bodylist))          ;more efficient than (= (length bodylist) 1)
        (car bodylist)
        bodylist)))


;;DESUGAR: <body in extended grammar> --> <body in kernel grammar>

(define (desugar body)
  (cond ((and? body)                    ;makes ugly output
                                        ;would be cleaner as a special form
         (let ((exps (expressions-of-and body)))
           (cond ((null? exps) #t)
                 ((null? (cdr exps)) (desugar (car exps)))
                 (else
                  `((lambda (x thunk)
                      (if x (thunk) #f))
                    ,(desugar (car exps))
                    (lambda () ,(desugar (make-and (cdr exps)))))))))
        ((or? body)                     ;likewise
         (let ((exps (expressions-of-or body)))
           (cond ((null? exps) #f)
                 ((null? (cdr exps)) (desugar (car exps)))
                 (else
                  `((lambda (x thunk)
                      (if x x (thunk)))
                    ,(desugar (car exps))
                    (lambda () ,(desugar (make-or (cdr exps)))))))))
        ((quote? body)
         (if (symbol-expression? body)
             body
             (let ((text (text-of-quote body)))
               (if (or (self-evaluating? text)
                       (submodel-null? text))
                   text
                   (make-combination
                    (cons 'list
                          (desugar
                           (map
                            (lambda (item)
                              (make-quote
                               item))
                            text))))))))
        ((begin? body)
         (make-begin
          (map desugar (expressions-of-begin body))))
        ((cond? body)
         (let ((clauses (clauses-of-cond body)))
           (if (null? clauses)
               submodel-useless-value
               (let ((first-clause (car clauses)))
                 (desugar
                  (if (else-clause? first-clause)
                      (expression-of-clause first-clause)
                      (make-if
                       (test-of-clause first-clause)
                       (expression-of-clause first-clause)
                       (make-cond (cdr clauses)))))))))
        ((let? body)
         (desugar
          (let ((bindings (bindings-of-let body)))
            (make-combination
             (cons
              (make-lambda
               (map variable-of-binding bindings)
               (body-of-let body))
              (map init-of-binding bindings))))))
        ((or (self-evaluating? body)
             (symbol-expression? body)
             (variable? body)
             (submodel-null? body))
         body)
        ((lambda-expression? body)
         (make-lambda
          (formals-of-lambda body)
          (desugar (body-of-lambda body))))
        ((combination? body)
         (make-combination
          (map desugar (expressions-of-combination body))))
        ((if? body)
         (make-if
          (desugar (test-of-if body))
          (desugar (consequent-of body))
          (if (no-alternative? body)
              submodel-useless-value
              (desugar (alternative-of body)))))
        ((set!? body)
         (make-set!
          (variable-of-set! body)
          (desugar (expression-of-set! body))))
        ((body-with-defines? body)
         (let ((defs (defines-of-body body)))
           (make-body
            (map make-define
                 (map variable-of-define defs)
                 (map desugar (map expression-of-define defs)))
            (desugar (expression-of-body body)))))
        (else (error "DESUGAR: unknown form: " body))))
